package org.apache.bcel.verifier.statics;

import java.util.Arrays;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantClass;
import org.apache.bcel.classfile.ConstantDouble;
import org.apache.bcel.classfile.ConstantFieldref;
import org.apache.bcel.classfile.ConstantFloat;
import org.apache.bcel.classfile.ConstantInteger;
import org.apache.bcel.classfile.ConstantInterfaceMethodref;
import org.apache.bcel.classfile.ConstantLong;
import org.apache.bcel.classfile.ConstantMethodref;
import org.apache.bcel.classfile.ConstantNameAndType;
import org.apache.bcel.classfile.ConstantString;
import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.LineNumber;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ALOAD;
import org.apache.bcel.generic.ANEWARRAY;
import org.apache.bcel.generic.ASTORE;
import org.apache.bcel.generic.ATHROW;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.BREAKPOINT;
import org.apache.bcel.generic.CHECKCAST;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.DLOAD;
import org.apache.bcel.generic.DSTORE;
import org.apache.bcel.generic.EmptyVisitor;
import org.apache.bcel.generic.FLOAD;
import org.apache.bcel.generic.FSTORE;
import org.apache.bcel.generic.FieldInstruction;
import org.apache.bcel.generic.GETSTATIC;
import org.apache.bcel.generic.GotoInstruction;
import org.apache.bcel.generic.IINC;
import org.apache.bcel.generic.ILOAD;
import org.apache.bcel.generic.IMPDEP1;
import org.apache.bcel.generic.IMPDEP2;
import org.apache.bcel.generic.INSTANCEOF;
import org.apache.bcel.generic.INVOKEDYNAMIC;
import org.apache.bcel.generic.INVOKEINTERFACE;
import org.apache.bcel.generic.INVOKESPECIAL;
import org.apache.bcel.generic.INVOKESTATIC;
import org.apache.bcel.generic.INVOKEVIRTUAL;
import org.apache.bcel.generic.ISTORE;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.LDC;
import org.apache.bcel.generic.LDC2_W;
import org.apache.bcel.generic.LLOAD;
import org.apache.bcel.generic.LOOKUPSWITCH;
import org.apache.bcel.generic.LSTORE;
import org.apache.bcel.generic.LoadClass;
import org.apache.bcel.generic.MULTIANEWARRAY;
import org.apache.bcel.generic.NEW;
import org.apache.bcel.generic.NEWARRAY;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.PUTSTATIC;
import org.apache.bcel.generic.RET;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.TABLESWITCH;
import org.apache.bcel.generic.Type;
import org.apache.bcel.verifier.PassVerifier;
import org.apache.bcel.verifier.VerificationResult;
import org.apache.bcel.verifier.Verifier;
import org.apache.bcel.verifier.VerifierFactory;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.ClassConstraintException;
import org.apache.bcel.verifier.exc.InvalidMethodException;
import org.apache.bcel.verifier.exc.StaticCodeConstraintException;
import org.apache.bcel.verifier.exc.StaticCodeInstructionConstraintException;
import org.apache.bcel.verifier.exc.StaticCodeInstructionOperandConstraintException;

/* loaded from: org_apache_bcel_verifier_statics.jar:org/apache/bcel/verifier/statics/Pass3aVerifier.class */
public final class Pass3aVerifier extends PassVerifier {
    private final Verifier myOwner;
    private final int methodNo;
    private InstructionList instructionList;
    private Code code;

    /* JADX INFO: Access modifiers changed from: private */
    /* loaded from: org_apache_bcel_verifier_statics.jar:org/apache/bcel/verifier/statics/Pass3aVerifier$InstOperandConstraintVisitor.class */
    public class InstOperandConstraintVisitor extends EmptyVisitor {
        private final ConstantPoolGen constantPoolGen;

        InstOperandConstraintVisitor(ConstantPoolGen constantPoolGen) {
            this.constantPoolGen = constantPoolGen;
        }

        private void constraintViolated(Instruction i, String message) {
            throw new StaticCodeInstructionOperandConstraintException("Instruction " + Pass3aVerifier.this.tostring(i) + " constraint violated: " + message);
        }

        private Method getMethod(JavaClass jc, InvokeInstruction invoke) {
            Method[] ms = jc.getMethods();
            for (Method element : ms) {
                if (element.getName().equals(invoke.getMethodName(this.constantPoolGen)) && Type.getReturnType(element.getSignature()).equals(invoke.getReturnType(this.constantPoolGen)) && Arrays.equals(Type.getArgumentTypes(element.getSignature()), invoke.getArgumentTypes(this.constantPoolGen))) {
                    return element;
                }
            }
            return null;
        }

        private Method getMethodRecursive(JavaClass jc, InvokeInstruction invoke) throws ClassNotFoundException {
            Method m = getMethod(jc, invoke);
            if (m != null) {
                return m;
            }
            for (JavaClass superclass : jc.getSuperClasses()) {
                Method m2 = getMethod(superclass, invoke);
                if (m2 != null) {
                    return m2;
                }
            }
            for (JavaClass superclass2 : jc.getInterfaces()) {
                Method m3 = getMethod(superclass2, invoke);
                if (m3 != null) {
                    return m3;
                }
            }
            return null;
        }

        private ObjectType getObjectType(FieldInstruction o) {
            ObjectType referenceType = o.getReferenceType(this.constantPoolGen);
            if (referenceType instanceof ObjectType) {
                return referenceType;
            }
            constraintViolated(o, "expecting ObjectType but got " + referenceType);
            return null;
        }

        private void indexValid(Instruction i, int idx) {
            if (idx < 0 || idx >= this.constantPoolGen.getSize()) {
                constraintViolated(i, "Illegal constant pool index '" + idx + "'.");
            }
        }

        private int maxLocals() {
            try {
                return Repository.lookupClass(Pass3aVerifier.this.myOwner.getClassName()).getMethods()[Pass3aVerifier.this.methodNo].getCode().getMaxLocals();
            } catch (ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }

        public void visitALOAD(ALOAD o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitANEWARRAY(ANEWARRAY o) {
            int dimensions;
            indexValid(o, o.getIndex());
            Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
            ArrayType type = o.getType(this.constantPoolGen);
            if ((type instanceof ArrayType) && (dimensions = type.getDimensions()) > 255) {
                constraintViolated(o, "Not allowed to create an array with more than 255 dimensions; actual: " + dimensions);
            }
        }

        public void visitASTORE(ASTORE o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitCHECKCAST(CHECKCAST o) {
            indexValid(o, o.getIndex());
            Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
        }

        public void visitDLOAD(DLOAD o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
                return;
            }
            int maxminus2 = maxLocals() - 2;
            if (idx > maxminus2) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
            }
        }

        public void visitDSTORE(DSTORE o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
                return;
            }
            int maxminus2 = maxLocals() - 2;
            if (idx > maxminus2) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
            }
        }

        public void visitFieldInstruction(FieldInstruction o) {
            try {
                indexValid(o, o.getIndex());
                Constant c = this.constantPoolGen.getConstant(o.getIndex());
                if (!(c instanceof ConstantFieldref)) {
                    constraintViolated(o, "Indexing a constant that's not a CONSTANT_Fieldref but a '" + Pass3aVerifier.this.tostring(c) + "'.");
                }
                String fieldName = o.getFieldName(this.constantPoolGen);
                JavaClass jc = Repository.lookupClass(getObjectType(o).getClassName());
                Field[] fields = jc.getFields();
                Field f = null;
                int length = fields.length;
                int i = 0;
                while (true) {
                    if (i >= length) {
                        break;
                    }
                    Field field = fields[i];
                    if (field.getName().equals(fieldName)) {
                        Type fType = Type.getType(field.getSignature());
                        Type oType = o.getType(this.constantPoolGen);
                        if (fType.equals(oType)) {
                            f = field;
                            break;
                        }
                    }
                    i++;
                }
                if (f == null) {
                    JavaClass[] superclasses = jc.getSuperClasses();
                    int length2 = superclasses.length;
                    int i2 = 0;
                    while (true) {
                        if (i2 >= length2) {
                            break;
                        }
                        JavaClass superclass = superclasses[i2];
                        Field[] fields2 = superclass.getFields();
                        for (Field field2 : fields2) {
                            if (field2.getName().equals(fieldName)) {
                                Type fType2 = Type.getType(field2.getSignature());
                                Type oType2 = o.getType(this.constantPoolGen);
                                if (fType2.equals(oType2)) {
                                    f = field2;
                                    if ((f.getAccessFlags() & 5) == 0) {
                                        f = null;
                                    }
                                }
                            }
                        }
                        i2++;
                    }
                    if (f == null) {
                        constraintViolated(o, "Referenced field '" + fieldName + "' does not exist in class '" + jc.getClassName() + "'.");
                    }
                } else {
                    Type.getType(f.getSignature());
                    o.getType(this.constantPoolGen);
                }
            } catch (ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }

        public void visitFLOAD(FLOAD o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitFSTORE(FSTORE o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitGETSTATIC(GETSTATIC o) {
            try {
                String fieldName = o.getFieldName(this.constantPoolGen);
                JavaClass jc = Repository.lookupClass(getObjectType(o).getClassName());
                Field[] fields = jc.getFields();
                Field f = null;
                int length = fields.length;
                int i = 0;
                while (true) {
                    if (i >= length) {
                        break;
                    }
                    Field field = fields[i];
                    if (!field.getName().equals(fieldName)) {
                        i++;
                    } else {
                        f = field;
                        break;
                    }
                }
                if (f == null) {
                    throw new AssertionViolatedException("Field '" + fieldName + "' not found in " + jc.getClassName());
                }
                if (!f.isStatic()) {
                    constraintViolated(o, "Referenced field '" + f + "' is not static which it should be.");
                }
            } catch (ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }

        public void visitIINC(IINC o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitILOAD(ILOAD o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitINSTANCEOF(INSTANCEOF o) {
            indexValid(o, o.getIndex());
            Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
        }

        public void visitINVOKEDYNAMIC(INVOKEDYNAMIC o) {
            throw new UnsupportedOperationException("INVOKEDYNAMIC instruction is not supported at this time");
        }

        public void visitInvokeInstruction(InvokeInstruction o) {
            indexValid(o, o.getIndex());
            if ((o instanceof INVOKEVIRTUAL) || (o instanceof INVOKESPECIAL) || (o instanceof INVOKESTATIC)) {
                ConstantMethodref constant = this.constantPoolGen.getConstant(o.getIndex());
                if (!(constant instanceof ConstantMethodref)) {
                    constraintViolated(o, "Indexing a constant that's not a CONSTANT_Methodref but a '" + Pass3aVerifier.this.tostring(constant) + "'.");
                } else {
                    ConstantNameAndType cnat = this.constantPoolGen.getConstant(constant.getNameAndTypeIndex());
                    ConstantUtf8 cutf8 = this.constantPoolGen.getConstant(cnat.getNameIndex());
                    if (cutf8.getBytes().equals("<init>") && !(o instanceof INVOKESPECIAL)) {
                        constraintViolated(o, "Only INVOKESPECIAL is allowed to invoke instance initialization methods.");
                    }
                    if (!cutf8.getBytes().equals("<init>") && cutf8.getBytes().startsWith("<")) {
                        constraintViolated(o, "No method with a name beginning with '<' other than the instance initialization methods may be called by the method invocation instructions.");
                    }
                }
            } else {
                ConstantInterfaceMethodref constant2 = this.constantPoolGen.getConstant(o.getIndex());
                if (!(constant2 instanceof ConstantInterfaceMethodref)) {
                    constraintViolated(o, "Indexing a constant that's not a CONSTANT_InterfaceMethodref but a '" + Pass3aVerifier.this.tostring(constant2) + "'.");
                }
                ConstantNameAndType cnat2 = this.constantPoolGen.getConstant(constant2.getNameAndTypeIndex());
                String name = this.constantPoolGen.getConstant(cnat2.getNameIndex()).getBytes();
                if (name.equals("<init>")) {
                    constraintViolated(o, "Method to invoke must not be '<init>'.");
                }
                if (name.equals("<clinit>")) {
                    constraintViolated(o, "Method to invoke must not be '<clinit>'.");
                }
            }
            Type t = o.getReturnType(this.constantPoolGen);
            if (t instanceof ArrayType) {
                t = ((ArrayType) t).getBasicType();
            }
            if (t instanceof ObjectType) {
                Verifier v = VerifierFactory.getVerifier(((ObjectType) t).getClassName());
                VerificationResult vr = v.doPass2();
                if (vr.getStatus() != 1) {
                    constraintViolated(o, "Return type class/interface could not be verified successfully: '" + vr.getMessage() + "'.");
                }
            }
            Type[] ts = o.getArgumentTypes(this.constantPoolGen);
            for (Type element : ts) {
                Type t2 = element;
                if (t2 instanceof ArrayType) {
                    t2 = ((ArrayType) t2).getBasicType();
                }
                if (t2 instanceof ObjectType) {
                    Verifier v2 = VerifierFactory.getVerifier(((ObjectType) t2).getClassName());
                    VerificationResult vr2 = v2.doPass2();
                    if (vr2.getStatus() != 1) {
                        constraintViolated(o, "Argument type class/interface could not be verified successfully: '" + vr2.getMessage() + "'.");
                    }
                }
            }
        }

        public void visitINVOKEINTERFACE(INVOKEINTERFACE o) {
            try {
                String className = o.getClassName(this.constantPoolGen);
                JavaClass jc = Repository.lookupClass(className);
                Method m = getMethodRecursive(jc, o);
                if (m == null) {
                    constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                }
                if (jc.isClass()) {
                    constraintViolated(o, "Referenced class '" + jc.getClassName() + "' is a class, but not an interface as expected.");
                }
            } catch (ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }

        public void visitINVOKESPECIAL(INVOKESPECIAL o) {
            try {
                String className = o.getClassName(this.constantPoolGen);
                JavaClass jc = Repository.lookupClass(className);
                Method m = getMethodRecursive(jc, o);
                if (m == null) {
                    constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                }
                JavaClass current = Repository.lookupClass(Pass3aVerifier.this.myOwner.getClassName());
                if (current.isSuper() && Repository.instanceOf(current, jc) && !current.equals(jc) && !o.getMethodName(this.constantPoolGen).equals("<init>")) {
                    int supidx = -1;
                    Method meth = null;
                    while (supidx != 0) {
                        supidx = current.getSuperclassNameIndex();
                        current = Repository.lookupClass(current.getSuperclassName());
                        Method[] meths = current.getMethods();
                        int length = meths.length;
                        int i = 0;
                        while (true) {
                            if (i >= length) {
                                break;
                            }
                            Method meth2 = meths[i];
                            if (!meth2.getName().equals(o.getMethodName(this.constantPoolGen)) || !Type.getReturnType(meth2.getSignature()).equals(o.getReturnType(this.constantPoolGen)) || !Arrays.equals(Type.getArgumentTypes(meth2.getSignature()), o.getArgumentTypes(this.constantPoolGen))) {
                                i++;
                            } else {
                                meth = meth2;
                                break;
                            }
                        }
                        if (meth != null) {
                            break;
                        }
                    }
                    if (meth == null) {
                        constraintViolated(o, "ACC_SUPER special lookup procedure not successful: method '" + o.getMethodName(this.constantPoolGen) + "' with proper signature not declared in superclass hierarchy.");
                    }
                }
            } catch (ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }

        public void visitINVOKESTATIC(INVOKESTATIC o) {
            try {
                String className = o.getClassName(this.constantPoolGen);
                JavaClass jc = Repository.lookupClass(className);
                Method m = getMethodRecursive(jc, o);
                if (m == null) {
                    constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                } else if (!m.isStatic()) {
                    constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' has ACC_STATIC unset.");
                }
            } catch (ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }

        public void visitINVOKEVIRTUAL(INVOKEVIRTUAL o) {
            try {
                String className = o.getClassName(this.constantPoolGen);
                JavaClass jc = Repository.lookupClass(className);
                Method m = getMethodRecursive(jc, o);
                if (m == null) {
                    constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                }
                if (!jc.isClass()) {
                    constraintViolated(o, "Referenced class '" + jc.getClassName() + "' is an interface, but not a class as expected.");
                }
            } catch (ClassNotFoundException e) {
                Pass3aVerifier.this.addMessage("Unable to verify INVOKEVITUAL as cannot load target class: " + e.getCause());
            }
        }

        public void visitISTORE(ISTORE o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitLDC(LDC ldc) {
            indexValid(ldc, ldc.getIndex());
            Constant c = this.constantPoolGen.getConstant(ldc.getIndex());
            if (c instanceof ConstantClass) {
                Pass3aVerifier.this.addMessage("Operand of LDC or LDC_W is CONSTANT_Class '" + Pass3aVerifier.this.tostring(c) + "' - this is only supported in JDK 1.5 and higher.");
            } else if (!(c instanceof ConstantInteger) && !(c instanceof ConstantFloat) && !(c instanceof ConstantString)) {
                constraintViolated(ldc, "Operand of LDC or LDC_W must be one of CONSTANT_Integer, CONSTANT_Float or CONSTANT_String, but is '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
        }

        public void visitLDC2_W(LDC2_W o) {
            indexValid(o, o.getIndex());
            Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantLong) && !(c instanceof ConstantDouble)) {
                constraintViolated(o, "Operand of LDC2_W must be CONSTANT_Long or CONSTANT_Double, but is '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
            try {
                indexValid(o, o.getIndex() + 1);
            } catch (StaticCodeInstructionOperandConstraintException e) {
                throw new AssertionViolatedException("Does not BCEL handle that? LDC2_W operand has a problem.", e);
            }
        }

        public void visitLLOAD(LLOAD o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
                return;
            }
            int maxminus2 = maxLocals() - 2;
            if (idx > maxminus2) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
            }
        }

        public void visitLoadClass(LoadClass loadClass) {
            ObjectType t = loadClass.getLoadClassType(this.constantPoolGen);
            if (t != null) {
                Verifier v = VerifierFactory.getVerifier(t.getClassName());
                VerificationResult vr = v.doPass1();
                if (vr.getStatus() != 1) {
                    constraintViolated((Instruction) loadClass, "Class '" + loadClass.getLoadClassType(this.constantPoolGen).getClassName() + "' is referenced, but cannot be loaded: '" + vr + "'.");
                }
            }
        }

        public void visitLOOKUPSWITCH(LOOKUPSWITCH o) {
            int[] matchs = o.getMatchs();
            int max = Integer.MIN_VALUE;
            for (int i = 0; i < matchs.length; i++) {
                if (matchs[i] == max && i != 0) {
                    constraintViolated(o, "Match '" + matchs[i] + "' occurs more than once.");
                }
                if (matchs[i] < max) {
                    constraintViolated(o, "Lookup table must be sorted but isn't.");
                } else {
                    max = matchs[i];
                }
            }
        }

        public void visitLSTORE(LSTORE o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
                return;
            }
            int maxminus2 = maxLocals() - 2;
            if (idx > maxminus2) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
            }
        }

        public void visitMULTIANEWARRAY(MULTIANEWARRAY o) {
            indexValid(o, o.getIndex());
            Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
            int dimensions2create = o.getDimensions();
            if (dimensions2create < 1) {
                constraintViolated(o, "Number of dimensions to create must be greater than zero.");
            }
            ArrayType type = o.getType(this.constantPoolGen);
            if (type instanceof ArrayType) {
                int dimensions = type.getDimensions();
                if (dimensions < dimensions2create) {
                    constraintViolated(o, "Not allowed to create array with more dimensions ('" + dimensions2create + "') than the one referenced by the CONSTANT_Class '" + type + "'.");
                    return;
                }
                return;
            }
            constraintViolated(o, "Expecting a CONSTANT_Class referencing an array type. [Constraint not found in The Java Virtual Machine Specification, Second Edition, 4.8.1]");
        }

        public void visitNEW(NEW o) {
            indexValid(o, o.getIndex());
            ConstantClass constant = this.constantPoolGen.getConstant(o.getIndex());
            if (!(constant instanceof ConstantClass)) {
                constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(constant) + "'.");
                return;
            }
            ConstantUtf8 cutf8 = this.constantPoolGen.getConstant(constant.getNameIndex());
            Type t = Type.getType("L" + cutf8.getBytes() + ";");
            if (t instanceof ArrayType) {
                constraintViolated(o, "NEW must not be used to create an array.");
            }
        }

        public void visitNEWARRAY(NEWARRAY o) {
            byte t = o.getTypecode();
            if (t != 4 && t != 5 && t != 6 && t != 7 && t != 8 && t != 9 && t != 10 && t != 11) {
                constraintViolated(o, "Illegal type code '" + Pass3aVerifier.this.tostring(Byte.valueOf(t)) + "' for 'atype' operand.");
            }
        }

        public void visitPUTSTATIC(PUTSTATIC o) {
            try {
                String fieldName = o.getFieldName(this.constantPoolGen);
                JavaClass jc = Repository.lookupClass(getObjectType(o).getClassName());
                Field[] fields = jc.getFields();
                Field f = null;
                int length = fields.length;
                int i = 0;
                while (true) {
                    if (i >= length) {
                        break;
                    }
                    Field field = fields[i];
                    if (!field.getName().equals(fieldName)) {
                        i++;
                    } else {
                        f = field;
                        break;
                    }
                }
                if (f == null) {
                    throw new AssertionViolatedException("Field '" + fieldName + "' not found in " + jc.getClassName());
                }
                if (f.isFinal() && !Pass3aVerifier.this.myOwner.getClassName().equals(getObjectType(o).getClassName())) {
                    constraintViolated(o, "Referenced field '" + f + "' is final and must therefore be declared in the current class '" + Pass3aVerifier.this.myOwner.getClassName() + "' which is not the case: it is declared in '" + o.getReferenceType(this.constantPoolGen) + "'.");
                }
                if (!f.isStatic()) {
                    constraintViolated(o, "Referenced field '" + f + "' is not static which it should be.");
                }
                String methName = Repository.lookupClass(Pass3aVerifier.this.myOwner.getClassName()).getMethods()[Pass3aVerifier.this.methodNo].getName();
                if (!jc.isClass() && !methName.equals("<clinit>")) {
                    constraintViolated(o, "Interface field '" + f + "' must be set in a '<clinit>' method.");
                }
            } catch (ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }

        public void visitRET(RET o) {
            int idx = o.getIndex();
            if (idx < 0) {
                constraintViolated(o, "Index '" + idx + "' must be non-negative.");
                return;
            }
            int maxminus1 = maxLocals() - 1;
            if (idx > maxminus1) {
                constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
            }
        }

        public void visitTABLESWITCH(TABLESWITCH o) {
        }
    }

    private static boolean contains(int[] ints, int i) {
        for (int k : ints) {
            if (k == i) {
                return true;
            }
        }
        return false;
    }

    public Pass3aVerifier(Verifier owner, int methodNo) {
        this.myOwner = owner;
        this.methodNo = methodNo;
    }

    private void delayedPass2Checks() {
        int[] instructionPositions = this.instructionList.getInstructionPositions();
        int codeLength = this.code.getCode().length;
        LineNumberTable lnt = this.code.getLineNumberTable();
        if (lnt != null) {
            LineNumber[] lineNumbers = lnt.getLineNumberTable();
            IntList offsets = new IntList();
            for (LineNumber lineNumber : lineNumbers) {
                for (int instructionPosition : instructionPositions) {
                    int offset = lineNumber.getStartPC();
                    if (instructionPosition == offset) {
                        if (offsets.contains(offset)) {
                            addMessage("LineNumberTable attribute '" + this.code.getLineNumberTable() + "' refers to the same code offset ('" + offset + "') more than once which is violating the semantics [but is sometimes produced by IBM's 'jikes' compiler].");
                        } else {
                            offsets.add(offset);
                        }
                    }
                }
                throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has a LineNumberTable attribute '" + this.code.getLineNumberTable() + "' referring to a code offset ('" + lineNumber.getStartPC() + "') that does not exist.");
            }
        }
        for (LocalVariableTable localVariableTable : this.code.getAttributes()) {
            if (localVariableTable instanceof LocalVariableTable) {
                localVariableTable.forEach(localVariable -> {
                    int startpc = localVariable.getStartPC();
                    int length = localVariable.getLength();
                    if (!contains(instructionPositions, startpc)) {
                        throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has a LocalVariableTable attribute '" + this.code.getLocalVariableTable() + "' referring to a code offset ('" + startpc + "') that does not exist.");
                    }
                    if (!contains(instructionPositions, startpc + length) && startpc + length != codeLength) {
                        throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has a LocalVariableTable attribute '" + this.code.getLocalVariableTable() + "' referring to a code offset start_pc+length ('" + (startpc + length) + "') that does not exist.");
                    }
                });
            }
        }
        CodeException[] exceptionTable = this.code.getExceptionTable();
        for (CodeException element : exceptionTable) {
            int startpc = element.getStartPC();
            int endpc = element.getEndPC();
            int handlerpc = element.getHandlerPC();
            if (startpc >= endpc) {
                throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has its start_pc ('" + startpc + "') not smaller than its end_pc ('" + endpc + "').");
            }
            if (!contains(instructionPositions, startpc)) {
                throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its start_pc ('" + startpc + "').");
            }
            if (!contains(instructionPositions, endpc) && endpc != codeLength) {
                throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its end_pc ('" + startpc + "') [that is also not equal to code_length ('" + codeLength + "')].");
            }
            if (!contains(instructionPositions, handlerpc)) {
                throw new ClassConstraintException("Code attribute '" + tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its handler_pc ('" + handlerpc + "').");
            }
        }
    }

    public VerificationResult do_verify() {
        try {
            if (this.myOwner.doPass2().equals(VerificationResult.VR_OK)) {
                JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
                Method[] methods = jc.getMethods();
                if (this.methodNo >= methods.length) {
                    throw new InvalidMethodException("METHOD DOES NOT EXIST!");
                }
                Method method = methods[this.methodNo];
                this.code = method.getCode();
                if (method.isAbstract() || method.isNative()) {
                    return VerificationResult.VR_OK;
                }
                try {
                    this.instructionList = new InstructionList(method.getCode().getCode());
                    this.instructionList.setPositions(true);
                    VerificationResult vr = VerificationResult.VR_OK;
                    try {
                        delayedPass2Checks();
                        try {
                            pass3StaticInstructionChecks();
                            pass3StaticInstructionOperandsChecks();
                        } catch (ClassCastException cce) {
                            vr = new VerificationResult(2, "Class Cast Exception: " + cce.getMessage());
                        } catch (StaticCodeConstraintException | ClassFormatException e) {
                            vr = new VerificationResult(2, e.getMessage());
                        }
                        return vr;
                    } catch (ClassConstraintException | ClassFormatException e2) {
                        VerificationResult vr2 = new VerificationResult(2, e2.getMessage());
                        return vr2;
                    }
                } catch (RuntimeException e3) {
                    return new VerificationResult(2, "Bad bytecode in the code array of the Code attribute of method '" + tostring(method) + "'.");
                }
            }
            return VerificationResult.VR_NOTYET;
        } catch (ClassNotFoundException e4) {
            throw new AssertionViolatedException("Missing class: " + e4, e4);
        }
    }

    public int getMethodNo() {
        return this.methodNo;
    }

    private void pass3StaticInstructionChecks() {
        if (this.code.getCode().length >= 65536) {
            throw new StaticCodeInstructionConstraintException("Code array in code attribute '" + tostring(this.code) + "' too big: must be smaller than 6553665536 bytes.");
        }
        InstructionHandle start = this.instructionList.getStart();
        while (true) {
            InstructionHandle ih = start;
            if (ih != null) {
                Instruction i = ih.getInstruction();
                if (i instanceof IMPDEP1) {
                    throw new StaticCodeInstructionConstraintException("IMPDEP1 must not be in the code, it is an illegal instruction for _internal_ JVM use!");
                }
                if (i instanceof IMPDEP2) {
                    throw new StaticCodeInstructionConstraintException("IMPDEP2 must not be in the code, it is an illegal instruction for _internal_ JVM use!");
                }
                if (i instanceof BREAKPOINT) {
                    throw new StaticCodeInstructionConstraintException("BREAKPOINT must not be in the code, it is an illegal instruction for _internal_ JVM use!");
                }
                start = ih.getNext();
            } else {
                Instruction last = this.instructionList.getEnd().getInstruction();
                if (!(last instanceof ReturnInstruction) && !(last instanceof RET) && !(last instanceof GotoInstruction) && !(last instanceof ATHROW)) {
                    throw new StaticCodeInstructionConstraintException("Execution must not fall off the bottom of the code array. This constraint is enforced statically as some existing verifiers do - so it may be a false alarm if the last instruction is not reachable.");
                }
                return;
            }
        }
    }

    private void pass3StaticInstructionOperandsChecks() {
        try {
            ConstantPoolGen cpg = new ConstantPoolGen(Repository.lookupClass(this.myOwner.getClassName()).getConstantPool());
            InstOperandConstraintVisitor v = new InstOperandConstraintVisitor(cpg);
            for (InstructionHandle ih = this.instructionList.getStart(); ih != null; ih = ih.getNext()) {
                JsrInstruction instruction = ih.getInstruction();
                if (instruction instanceof JsrInstruction) {
                    InstructionHandle target = instruction.getTarget();
                    if (target == this.instructionList.getStart()) {
                        throw new StaticCodeInstructionOperandConstraintException("Due to JustIce's clear definition of subroutines, no JSR or JSR_W may have a top-level instruction (such as the very first instruction, which is targeted by instruction '" + tostring(ih) + "' as its target.");
                    }
                    if (!(target.getInstruction() instanceof ASTORE)) {
                        throw new StaticCodeInstructionOperandConstraintException("Due to JustIce's clear definition of subroutines, no JSR or JSR_W may target anything else than an ASTORE instruction. Instruction '" + tostring(ih) + "' targets '" + tostring(target) + "'.");
                    }
                }
                ih.accept(v);
            }
        } catch (ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }

    protected String tostring(Object obj) {
        String ret;
        try {
            ret = obj.toString();
        } catch (RuntimeException e) {
            String s = obj.getClass().getName();
            ret = "<<" + s.substring(s.lastIndexOf(".") + 1) + ">>";
        }
        return ret;
    }
}
